;Clock with roman numerals (hours and minutes)
;ATTiny2313 @ 4.194304 MHz (external XTAL)
;
;DISPLAY
;  2 halves, multiplexed, hours: 5 digits, minutes; 7 digits
;  digits enabled at the same time: 1 and 7, 2 and 8, 3 and 9, ..., 6 and 12
;  shift register: 16 bit per display half, shift in "a1" segment last (output Q0)
;
;      a1   a2
;      === ===
;     |\h |ij/|
;   f | \ | / | b
;     |g1\|/g2|
;      === ===
;     |  /|\  |
;   e | / | \ | c
;     |/k |lm\|
;      === ===
;      d1   d2
;
;PORTS
;  PD0 = Data for displays 1..6 (left)
;  PD1 = Data for displays 7..12 (right)
;  PD2 = Clock for all displays
;  PD3 = Strobe for all displays
;  PD4 = 12/24h mode (low = 12h)
;  PD5 = UTC/local time (low = UTC)
;  PD6 = DCF77 input (Timer1 ICP)
;  PB0 = Displays 1 and 7 Anode (active low)
;  PB1 = Displays 2 and 8 Anode (active low)
;  PB2 = Displays 3 and 9 Anode (active low)
;  PB3 = Displays 4 and 10 Anode (active low)
;  PB4 = Displays 5 and 11 Anode (active low)
;  PB5 = Displays 6 and 12 Anode (active low)
;  PB6 = Colon (active low)
;  PB7 = Display Mode (open = high = roman, low = binary)
;
;VERSIONS
;  2009-07-05  Arne Rossius
;    * first version, just multiplexing with static display
;  2010-02-06
;    * added DCF77 receiver with plausibility check
;  2010-02-26  Arne Rossius
;    * replaced direct RAM offsets by RAM offset constants

;TODO: display "2A" and "2B" hours when disabling DST

.equ	DCF_INVERTED = 1 ;inverted DCF77 signal (low pulses)
.equ	DCF_PULLUP = 1 ;internal pull-up resistor for DCF77 input

;===============================================================================

.include "../include/tn2313def.inc"

.equ	ANODE_PORT = PORTB
.equ	SHIFT_PORT = PORTD
.equ	SHIFT_DATA1 = 0
.equ	SHIFT_DATA2 = 1
.equ	SHIFT_CLOCK = 2
.equ	SHIFT_STROBE = 3
.equ	DISPLAY_PIN = PINB
.equ	DISPLAY_MODE = 7
.equ	SETTINGS_PIN = PIND
.equ	SET_12H = 4
.equ	SET_UTC = 5

.def	byte1 = R0
.def	byte2 = R1
.def	byte3 = R2
.def	byte4 = R3
.def	seconds = R4
.def	intdiv = R5

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	display = R19
.def	dcfbits = R20
.def	dcfflags = R21

;R29:R28 = Y register (RAM pointer)
;R31:R30 = Z register (Flash pointer)

;===============================================================================

.dseg
	RAM_Display: .byte 12
	
;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org OVF0addr
	rjmp	int_16hz
	
.equ DCF_USE_INT0 = 0
.include "dcf77.asm"

;===============================================================================
	
charset:
	.include "charset.asm"
	
init_string:
	;string to display until valid time has been received
	.db "ROMAN", "CLOCK  "
	
hours:
	;display data for hours (roman numerals 0 to 23, right-aligned)
	.db "    -", "    I", "   II", "  III", "   IV", "    V"
	.db "   VI", "  VII", " VIII", "   IX", "    X", "   XI"
	.db "  XII", " XIII", "  XIV", "   XV", "  XVI", " XVII"
	.db "XVIII", "  XIX", "   XX", "  XXI", " XXII", "XXIII"
minutes:
	;display data for minutes (roman numerals 0 to 59, left-aligned)
	.db "-      ", "I      ", "II     ", "III    ", "IV     ", "V      "
	.db "VI     ", "VII    ", "VIII   ", "IX     ", "X      ", "XI     "
	.db "XII    ", "XIII   ", "XIV    ", "XV     ", "XVI    ", "XVII   "
	.db "XVIII  ", "XIX    ", "XX     ", "XXI    ", "XXII   ", "XXIII  "
	.db "XXIV   ", "XXV    ", "XXVI   ", "XXVII  ", "XXVIII ", "XXIX   "
	.db "XXX    ", "XXXI   ", "XXXII  ", "XXXIII ", "XXXIV  ", "XXXV   "
	.db "XXXVI  ", "XXXVII ", "XXXVIII", "XXXIX  ", "XL     ", "XLI    "
	.db "XLII   ", "XLIII  ", "XLIV   ", "XLV    ", "XLVI   ", "XLVII  "
	.db "XLVIII ", "XLIX   ", "L      ", "LI     ", "LII    ", "LIII   "
	.db "LIV    ", "LV     ", "LVI    ", "LVII   ", "LVIII  ", "LIX    "
	
;===============================================================================

int_16hz:
	;16 Hz periodic interrupt
	push	temp
	push	temp2
	push	YL
	push	ZL
	push	ZH
	in	temp, SREG
	push	temp
	
	;divide to 1 Hz
	inc	intdiv
	ldi	temp, 0x0F
	and	intdiv, temp
	breq	PC+2
	rjmp	int_16hz_end
	
	;increment time
	sbrs	dcfflags, dcfTimeOK ;don't increment when no valid time is set yet
	rjmp	int_1hz_inctime_end
	mov	temp, seconds
	rcall	bcdinc
	mov	seconds, temp
	cpi	temp, 0x60
	brlo	int_1hz_inctime_end
	clr	seconds
	ldi	YL, RAM_Time
	rcall	inctime
	rcall	calculate_utc
int_1hz_inctime_end:
	
	;update display
	sbrs	dcfflags, dcfTimeOK ;don't update when no valid time is set yet
	rjmp	int_1hz_update_end
	
	lds	temp, RAM_Time+tHour ;hours
	sbis	SETTINGS_PIN, SET_UTC
	lds	temp, RAM_UTC+tHour
	rcall	bcd2bin
	sbic	SETTINGS_PIN, SET_12H
	rjmp	int_1hz_12h_end
	;convert to 12-hour format
	cpi	temp, 13
	brlo	PC+2
	subi	temp, 12
	cpi	temp, 0
	brne	PC+2
	ldi	temp, 12
int_1hz_12h_end:
	ldi	YL, RAM_Display
	sbic	DISPLAY_PIN, DISPLAY_MODE
	rjmp	int_1hz_hour_roman
	;Binary Hours
	ldi	temp2, ' ' ;Bit 4
	sbrc	temp, 4
	ldi	temp2, '1'
	st	Y+, temp2
	sbrc	temp2, 4 ;Bit 3
	ldi	temp2, '0'
	sbrc	temp, 3
	ldi	temp2, '1'
	st	Y+, temp2
	sbrc	temp2, 4 ;Bit 2
	ldi	temp2, '0'
	sbrc	temp, 2
	ldi	temp2, '1'
	st	Y+, temp2
	sbrc	temp2, 4 ;Bit 1
	ldi	temp2, '0'
	sbrc	temp, 1
	ldi	temp2, '1'
	st	Y+, temp2
	ldi	temp2, '0' ;Bit 0
	sbrc	temp, 0
	ldi	temp2, '1'
	st	Y+, temp2
	rjmp	int_1hz_hour_end
int_1hz_hour_roman:
	;Roman Hours
	ldi	ZL, LOW(hours*2)
	ldi	ZH, HIGH(hours*2)
	ldi	temp2, 0
	add	ZL, temp
	adc	ZH, temp2
	lsl	temp
	lsl	temp
	add	ZL, temp
	adc	ZH, temp2
	ldi	temp2, 5
int_1hz_update_h:
	lpm	temp, Z+
	st	Y+, temp
	dec	temp2
	brne	int_1hz_update_h
int_1hz_hour_end:
	
	lds	temp, RAM_Time+tMinute ;minutes
	sbis	SETTINGS_PIN, SET_UTC
	lds	temp, RAM_UTC+tMinute
	rcall	bcd2bin
	sbic	DISPLAY_PIN, DISPLAY_MODE
	rjmp	int_1hz_minute_roman
	;Binary Minute
	cpi	temp, 32 ;Bit 5
	brlo	PC+3
	ldi	temp2, '1'
	st	Y+, temp2
	cpi	temp, 16 ;Bit 4
	brlo	PC+5
	ldi	temp2, '0'
	sbrc	temp, 4
	ldi	temp2, '1'
	st	Y+, temp2
	cpi	temp, 8 ;Bit 3
	brlo	PC+5
	ldi	temp2, '0'
	sbrc	temp, 3
	ldi	temp2, '1'
	st	Y+, temp2
	cpi	temp, 4 ;Bit 2
	brlo	PC+5
	ldi	temp2, '0'
	sbrc	temp, 2
	ldi	temp2, '1'
	st	Y+, temp2
	cpi	temp, 2 ;Bit 1
	brlo	PC+5
	ldi	temp2, '0'
	sbrc	temp, 1
	ldi	temp2, '1'
	st	Y+, temp2
	ldi	temp2, '0'
	sbrc	temp, 0
	ldi	temp2, '1'
	st	Y+, temp2
int_1hz_minute_fill:
	cpi	YL, RAM_Display+12
	breq	int_1hz_minute_end
	ldi	temp, ' '
	st	Y+, temp
	rjmp	int_1hz_minute_fill
int_1hz_minute_roman:
	;Roman Minute
	ldi	ZL, LOW(minutes*2)
	ldi	ZH, HIGH(minutes*2)
	ldi	temp2, 0
	add	ZL, temp
	adc	ZH, temp2
	lsl	temp
	add	ZL, temp
	adc	ZH, temp2
	lsl	temp
	add	ZL, temp
	adc	ZH, temp2
	ldi	temp2, 7
int_1hz_update_m:
	lpm	temp, Z+
	st	Y+, temp
	dec	temp2
	brne	int_1hz_update_m
int_1hz_minute_end:

int_1hz_update_end:
	
int_16hz_end:
	pop	temp
	out	SREG, temp
	pop	ZH
	pop	ZL
	pop	YL
	pop	temp2
	pop	temp
	reti

;===============================================================================

reset:
	;set stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;set ports
	ldi	temp, 0x7F
	out	DDRB, temp
	ldi	temp, 0xFF
	out	PORTB, temp
	ldi	temp, 0x0F
	out	DDRD, temp
	.if (DCF_PULLUP == 0)
	ldi	temp, 0x30
	.else
	ldi	temp, 0x70
	.endif
	out	PORTD, temp
	
	;init timer 0 for seconds increment/display update
	ldi	temp, 0x05 ;Clk/1024 => 16 Hz overflow @ 4.1943043 MHz
	out	TCCR0B, temp
	;enable timer 0 overflow int.
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;init registers
	clr	seconds
	clr	intdiv
	
	;init DCF77 reception
	rcall	dcf_init
	
	ldi	ZL, LOW(init_string*2)
	ldi	ZH, HIGH(init_string*2)
	ldi	YL, RAM_Display
	ldi	temp2, 12
init_ram:
	lpm	temp, Z+
	st	Y+, temp
	dec	temp2
	brne	init_ram
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------

loop:
	;multiplex display
	ldi	YL, RAM_Display
	ldi	display, 0xFE
mux_loop:
	;get chars from RAM and read display data from charset
	ldd	temp, Y+6 ;get data for right display half
	sbrc	dcfflags, dcfTimeOK ;display DCF77 reception information until time is set
	rjmp	mux_dcfinfo_end
	cpi	YL, RAM_Display+4; next-to-last digit: DCF77 input state
	brne	mux_dcfinput_end
	ldi	temp, '_'
	in	temp2, TCCR1B
	.if (DCF_INVERTED == 0)
	sbrs	temp2, ICES1
	.else
	sbrc	temp2, ICES1
	.endif
	ldi	temp, '-'
mux_dcfinput_end:
	cpi	YL, RAM_Display+5; last digit: DCF77 error (if any)
	brne	mux_dcferror_end
	ldi	temp, '/'
	sbrc	dcfflags, dcfFirstTelegram
	ldi	temp, '1' ;1 = 1 telegram received sucessfully
	sbrc	dcfflags, dcfStop
	ldi	temp, 'W' ;W = Waiting (DCF stopped, but no error)
	mov	temp2, dcfflags
	andi	temp2, dcfErrMask
	cpi	temp2, dcfErrBitLength
	brne	PC+2
	ldi	temp, 'L' ;L = Bit Length error
	cpi	temp2, dcfErrBitCount
	brne	PC+2
	ldi	temp, 'C' ;C = Bit Count error
	cpi	temp2, dcfErrParity
	brne	PC+2
	ldi	temp, 'P' ;P = Parity error
	cpi	temp2, dcfErrStartBit
	brne	PC+2
	ldi	temp, 'S' ;S = Start Bit error
	cpi	temp2, dcfErrDST
	brne	PC+2
	ldi	temp, 'D' ;D = invalid DST bits
	cpi	temp2, dcfErrPlausibility
	brne	PC+2
	ldi	temp, '!' ;! = Plausibility Check failed
mux_dcferror_end:
mux_dcfinfo_end:
	ldi	temp2, 0
	ldi	ZL, LOW(charset*2 - 0x20*2)
	ldi	ZH, HIGH(charset*2 - 0x20*2)
	lsl	temp
	add	ZL, temp
	adc	ZH, temp2
	lpm	byte3, Z+
	lpm	byte4, Z
	ld	temp, Y+ ;get data for left display half
	ldi	ZL, LOW(charset*2 - 0x20*2)
	ldi	ZH, HIGH(charset*2 - 0x20*2)
	lsl	temp
	add	ZL, temp
	adc	ZH, temp2
	lpm	byte1, Z+
	lpm	byte2, Z
	;shift display data into display shift registers
	ldi	temp, 16
mux_shift:
	cbi	SHIFT_PORT, SHIFT_DATA1
	cbi	SHIFT_PORT, SHIFT_DATA2
	lsl	byte1
	rol	byte2
	brcc	PC+2
	sbi	SHIFT_PORT, SHIFT_DATA1
	lsl	byte3
	rol	byte4
	brcc	PC+2
	sbi	SHIFT_PORT, SHIFT_DATA2
	sbi	SHIFT_PORT, SHIFT_CLOCK
	cbi	SHIFT_PORT, SHIFT_CLOCK
	dec	temp
	brne	mux_shift
	mov	temp2, display
	sbrc	dcfflags, dcfTimeOK
	andi	temp2, ~0x40 ;enable colon when displaying time
	;switch to next display and output new data
	ldi	temp, 0xFF ;all displays off
	out	ANODE_PORT, temp
	sbi	SHIFT_PORT, SHIFT_STROBE
	cbi	SHIFT_PORT, SHIFT_STROBE
	out	ANODE_PORT, temp2 ;enable display
	;short delay
	ldi	temp, 5
	ldi	temp2, 0
mux_delay:
	dec	temp2
	brne	mux_delay
	dec	temp
	brne	mux_delay
	;next display
	sec
	rol	display
	sbrc	display, 6
	rjmp	mux_loop
	rjmp	loop

;===============================================================================

bcd2bin:
	;convert packed BCD value to binary
	mov	temp2, temp
	andi	temp, 0x0F
	push	temp
	swap	temp2
	andi	temp2, 0x0F
	lsl	temp2
	mov	temp, temp2
	lsl	temp2
	lsl	temp2
	add	temp2, temp
	pop	temp
	add	temp, temp2
	ret
	